import os
from asyncio import Protocol, Future, Transport
from enum import auto, Enum
from typing import Optional, Tuple, List, Dict, Union

import attr
from h2.connection import H2Connection
from h2.events import (
    ResponseReceived, DataReceived, StreamEnded,
    StreamReset,
)
from hyperframe.frame import SettingsFrame
from structlog import wrap_logger, PrintLogger, BoundLogger

from aapns.errors import Disconnected
from . import errors


SIZE = 4096


@attr.s
class PendingResponse:
    logger = attr.ib()
    future = attr.ib(default=attr.Factory(Future))
    headers = attr.ib(default=None)
    body = attr.ib(default=b'')

    def to_response(self):
        headers = {
            key.decode('utf-8'): value.decode('utf-8')
            for key, value in self.headers
        }
        status = int(headers[':status'])
        return Response(status, headers, self.body)


@attr.s
class Response:
    status: int = attr.ib()
    headers: Dict[str, str] = attr.ib()
    body: bytes = attr.ib()


class States(Enum):
    connecting = auto()
    connected = auto()
    disconnected = auto()


class APNSProtocol(Protocol):
    def __init__(self, authority: str, logger: Optional[BoundLogger]=None):
        self.authority = authority
        self.logger = logger or wrap_logger(PrintLogger(open(os.devnull, 'w')))
        self.conn = H2Connection()
        self.transport: Union[Transport, None] = None
        self.responses: Dict[int, PendingResponse] = {}
        self.state = States.connecting

    async def request(self,
                      headers: List[Tuple[str, str]],
                      body: bytes) -> Response:
        if self.state is States.disconnected:
            raise Disconnected()
        stream_id = self.conn.get_next_available_stream_id()
        logger = self.logger.bind(stream_id=stream_id)
        pending = self.responses[stream_id] = PendingResponse(logger=logger)
        logger.debug('request', headers=headers, body=body)
        self.conn.send_headers(stream_id, headers)
        self.conn.send_data(stream_id, body, end_stream=True)
        if self.transport is not None:
            data_to_send = self.conn.data_to_send()
            if data_to_send:
                self.transport.write(data_to_send)
        await pending.future
        return pending.to_response()

    async def close(self):
        self.conn.close_connection()
        if self.transport:
            self.transport.write(self.conn.data_to_send())
            self.transport.close()
        self.transport = None
        self.state = States.disconnected

    def connection_made(self, transport: Transport):
        self.logger.debug('connected')
        self.transport = transport
        self.conn.initiate_connection()

        # This reproduces the error in #396, by changing the header table size.
        self.conn.update_settings({SettingsFrame.HEADER_TABLE_SIZE: SIZE})

        self.transport.write(self.conn.data_to_send())
        self.state = States.connected

    def connection_lost(self, exc):
        self.logger.debug('disconnected')
        self.transport = None
        for pending in self.responses.values():
            pending.future.set_exception(Disconnected())
        self.state = States.disconnected

    def data_received(self, data: bytes):
        events = self.conn.receive_data(data)

        for event in events:
            if isinstance(event, ResponseReceived):
                self.handle_response(event.headers, event.stream_id)
            elif isinstance(event, DataReceived):
                self.handle_data(event.data, event.stream_id)
            elif isinstance(event, StreamEnded):
                self.end_stream(event.stream_id)
            elif isinstance(event, StreamReset):
                self.reset_stream(event.stream_id)
            else:
                self.logger.debug('ignored', h2event=event)

        data = self.conn.data_to_send()
        if data:
            self.transport.write(data)

    def handle_response(self, response_headers: List[Tuple[bytes, bytes]], stream_id: int):
        if stream_id in self.responses:
            self.responses[stream_id].logger.debug(
                'response-headers',
                headers=response_headers
            )
            self.responses[stream_id].headers = response_headers
        else:
            self.logger.warning(
                'unexpected-response',
                stream_id=stream_id,
                headers=response_headers
            )

    def handle_data(self, data: bytes, stream_id: int):
        if stream_id in self.responses:
            self.responses[stream_id].logger.debug(
                'response-body',
                data=data
            )
            self.responses[stream_id].body += data
        else:
            self.logger.warning(
                'unexpected-data',
                stream_id=stream_id,
                data=data
            )

    def end_stream(self, stream_id: int):
        if stream_id in self.responses:
            response = self.responses.pop(stream_id)
            response.logger.debug('end-stream')
            response.future.set_result(True)
        else:
            self.logger.warning('unexpected-end-stream', stream_id=stream_id)

    def reset_stream(self, stream_id: int):
        if stream_id in self.responses:
            response = self.responses.pop(stream_id)
            response.logger.debug('reset-stream')
            response.future.set_exception(errors.StreamResetError())
        else:
            self.logger.warning('unexpected-reset-stream', stream_id=stream_id)



from typing import Type, Dict


class APNSError(Exception):
    pass


class Disconnected(APNSError):
    pass


class StreamResetError(APNSError):
    pass


class ResponseError(APNSError):
    codename = None

    def __init__(self, reason: str, apns_id: str):
        self.reason = reason
        self.apns_id = apns_id
        super().__init__(reason)


class UnknownResponseError(ResponseError):
    codename = '!unknown'


CODES: Dict[str, Type[ResponseError]] = {}


def create(codename: str) -> Type[ResponseError]:
    cls: Type[ResponseError] = type(codename, (ResponseError,), {'codename': codename})
    CODES[codename] = cls
    return cls


BadCollapseId = create('BadCollapseId')
BadDeviceToken = create('BadDeviceToken')
BadExpirationDate = create('BadExpirationDate')
BadMessageId = create('BadMessageId')
BadPriority = create('BadPriority')
BadTopic = create('BadTopic')
DeviceTokenNotForTopic = create('DeviceTokenNotForTopic')
DuplicateHeaders = create('DuplicateHeaders')
IdleTimeout = create('IdleTimeout')
MissingDeviceToken = create('MissingDeviceToken')
MissingTopic = create('MissingTopic')
PayloadEmpty = create('PayloadEmpty')
BadCertificate = create('BadCertificate')
BadCertificateEnvironment = create('BadCertificateEnvironment')
ExpiredProviderToken = create('ExpiredProviderToken')
Forbidden = create('Forbidden')
InvalidProviderToken = create('InvalidProviderToken')
MissingProviderToken = create('MissingProviderToken')
BadPath = create('BadPath')
MethodNotAllowed = create('MethodNotAllowed')
Unregistered = create('Unregistered')
PayloadTooLarge = create('PayloadTooLarge')
TooManyProviderTokenUpdates = create('TooManyProviderTokenUpdates')
TooManyRequests = create('TooManyRequests')
InternalServerError = create('InternalServerError')
ServiceUnavailable = create('ServiceUnavailable')
Shutdown = create('Shutdown')


def get(reason: str, apns_id: str) -> ResponseError:
    return CODES.get(reason, UnknownResponseError)(reason, apns_id)



from asyncio import get_event_loop
from typing import Dict

import attr
import click
from structlog import get_logger

from . import config, connection, models


SERVERS: Dict[bool, Dict[bool, config.Server]] = {
    True: {
        True: config.ProductionAltPort,
        False: config.Production
    },
    False: {
        True: config.DevelopmentAltPort,
        False: config.Development
    }
}


@attr.s
class Context:
    token = attr.ib()
    cert = attr.ib()
    server = attr.ib()
    apns_id = attr.ib()
    expiration = attr.ib()
    priority = attr.ib()
    topic = attr.ib()
    collapse_id = attr.ib()
    verbose = attr.ib()


async def do_send(context: Context, notification: models.Notification) -> str:
    conn = await connection.connect(
        context.cert,
        context.server,
        logger=get_logger() if context.verbose else None
    )
    resp_id = await conn.send_notification(
        context.token,
        notification,
        apns_id=context.apns_id,
        expiration=context.expiration,
        priority=context.priority,
        topic=context.topic,
        collapse_id=context.collapse_id
    )
    await conn.close()
    return resp_id


def send(context: Context, notification: models.Notification):
    resp_id = get_event_loop().run_until_complete(do_send(context, notification))
    click.echo(resp_id)


@click.group()
@click.argument('token')
@click.option('--client-cert-path', envvar='CLIENT_CERT_PATH')
@click.option('--prod', is_flag=True, default=False)
@click.option('--alt-port', is_flag=True, default=False)
@click.option('--expiration', default=None)
@click.option('--immediately', is_flag=True, default=False)
@click.option('--topic', default=None)
@click.option('--collapse-id', default=None)
@click.option('--apns-id', default=None)
@click.option('--verbose', is_flag=True, default=False)
@click.pass_context
def main(ctx, token, client_cert_path, prod, alt_port, expiration, immediately, topic, collapse_id, apns_id, verbose):
    ctx.obj = Context(
        token=token,
        cert=client_cert_path,
        server=SERVERS[prod][alt_port],
        expiration=expiration,
        priority=config.Priority.immediately if immediately else config.Priority.normal,
        topic=topic,
        collapse_id=collapse_id,
        apns_id=apns_id,
        verbose=verbose,
    )


@main.command('simple')
@click.argument('body')
@click.option('--title', default=None)
@click.pass_context
def simple(ctx, title, body):
    notification = models.Notification(
        alert=models.Alert(
            title=title,
            body=body,
        )
    )
    send(ctx.obj, notification)


@main.command('localized')
@click.argument('body')
@click.option('--body-args', multiple=True)
@click.option('--title', default=None)
@click.option('--title-args', multiple=True)
@click.option('--badge', type=click.INT)
@click.pass_context
def localized(ctx, title, body, title_args, body_args, badge):
    notification = models.Notification(
        alert=models.Alert(
            body=models.Localized(body, list(body_args)),
            title=models.Localized(title, list(title_args)) if title else title,
        ),
        badge=badge,
    )
    send(ctx.obj, notification)



from enum import Enum

import attr


@attr.s
class Server:
    host = attr.ib()
    port = attr.ib()


Production = Server('api.push.apple.com', 443)
ProductionAltPort = attr.assoc(Production, port=2197)
Development = Server('api.development.push.apple.com', 443)
DevelopmentAltPort = attr.assoc(Development, port=2197)


class Priority(Enum):
    immediately = 10
    normal = 5



import json
import ssl
from asyncio import get_event_loop, wait_for, Lock, sleep
from typing import Optional

from structlog import BoundLogger

from . import errors, config, models, connection



class APNS:
    def __init__(self, client_cert_path, server, logger, ssl_context, *, auto_reconnect=False, timeout=None):
        self.client_cert_path = client_cert_path
        self.server = server
        self.logger = logger
        self.ssl_context = ssl_context
        self.auto_reconnect = auto_reconnect
        self.run_coroutine = self._run_with_timeout if timeout is None else self._run_with_timeout
        self.timeout = timeout
        self.connection = None
        self.connection_lock = Lock()

    async def send_notification(self,
                                token: str,
                                notification: models.Notification,
                                *,
                                apns_id: Optional[str]=None,
                                expiration: Optional[int]=None,
                                priority: config.Priority=config.Priority.normal,
                                topic: Optional[str]=None,
                                collapse_id: Optional[str]=None) -> str:
        conn = await self.run_coroutine(self.get_connection())

        request_body = notification.encode()
        request_headers = [
            (':method', 'POST'),
            (':authority', self.server.host),
            (':scheme', 'https'),
            (':path', f'/3/device/{token}'),
            ('content-length', str(len(request_body))),
            ('apns-priority', str(priority.value)),
        ]

        if apns_id:
            request_headers.append(('apns-id', apns_id))
        if expiration:
            request_headers.append(('apns-expiration', str(expiration)))
        if topic:
            request_headers.append(('apns-topic', topic))
        if collapse_id:
            request_headers.append(('apns-collapse-id', collapse_id))

        response = await self.run_coroutine(conn.request(
            headers=request_headers,
            body=request_body,
        ))

        response_id = response.headers.get('apns-id', '')

        if response.status != 200:
            try:
                reason = json.loads(response.body)['reason']
            except:
                reason = response.body
            exc = errors.get(reason, response_id)
            raise exc
        else:
            return response_id

    @classmethod
    async def init_with_connection(cls, client_cert_path, server, logger, ssl_context, *, auto_reconnect=False, timeout=None):
        apns = cls(client_cert_path, server, logger, ssl_context, auto_reconnect=auto_reconnect, timeout=timeout)
        await apns.run_coroutine(apns.connect())
        return apns

    async def close(self):
        if self.connected:
            await self.connection.close()

    @property
    def connected(self):
        return (
            self.connection is not None
            and
            self.connection.state is not connection.States.disconnected
        )

    async def _run_with_timeout(self, coro):
        return await wait_for(coro, self.timeout)

    @staticmethod
    async def _run_without_timeout(coro):
        return await coro

    async def connect(self):
        self.connection = connection.APNSProtocol(self.server.host, self.logger)
        await get_event_loop().create_connection(
            lambda: self.connection,
            self.server.host,
            self.server.port,
            ssl=self.ssl_context
        )

    async def get_connection(self) -> connection.APNSProtocol:
        async with self.connection_lock:
            if not self.connected:
                if not self.auto_reconnect:
                    raise errors.Disconnected()
                while True:
                    try:
                        await self.connect()
                        break
                    except ConnectionRefusedError:
                        await sleep(0.1)
            return self.connection


async def connect(client_cert_path: str,
                  server: config.Server,
                  *,
                  ssl_context: Optional[ssl.SSLContext]=None,
                  logger: Optional[BoundLogger]=None,
                  auto_reconnect: bool=False,
                  timeout: Optional[int]=None) -> APNS:
    if ssl_context is None:
        ssl_context: ssl.SSLContext = ssl.create_default_context()
        ssl_context.set_alpn_protocols(['h2'])
        ssl_context.set_npn_protocols(['h2'])
    ssl_context.load_cert_chain(client_cert_path)
    return await APNS.init_with_connection(
        client_cert_path=client_cert_path,
        server=server,
        ssl_context=ssl_context,
        logger=logger,
        auto_reconnect=auto_reconnect,
        timeout=timeout
    )



import json
from typing import Dict, Any, Union, List

import attr


def str_list(instance: object, attr: attr.Attribute, value: Any) -> None:
    if not isinstance(value, list):
        raise TypeError('Must be list of strings')
    for arg in value:
        if not isinstance(arg, str):
            raise TypeError('Must be list of strings')


@attr.s
class Localized:
    key = attr.ib(validator=attr.validators.instance_of(str))
    args = attr.ib(default=None, validator=attr.validators.optional(str_list))


def maybe_localized(thing: Union[str, Localized],
                    nonloc: str,
                    lockey: str,
                    locarg: str) -> Dict[str, Union[str, List[str]]]:
    if isinstance(thing, Localized):
        attr.validate(thing)
        localized = {
            lockey: thing.key
        }
        if thing.args:
            localized[locarg] = thing.args
        return localized
    else:
        return {nonloc: thing}


@attr.s
class Alert:
    body = attr.ib(validator=attr.validators.instance_of((str, Localized)))
    title = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of((str, Localized)))
    )
    action_loc_key = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    launch_image = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )

    def get_dict(self) -> Dict[str, Any]:
        attr.validate(self)
        alert = {}
        if self.title:
            alert.update(maybe_localized(
                self.title,
                'title',
                'title-loc-key',
                'title-loc-args'
            ))
        alert.update(maybe_localized(
            self.body,
            'body',
            'loc-key',
            'loc-args'
        ))
        if self.action_loc_key:
            alert['action-loc-key'] = self.action_loc_key
        if self.launch_image:
            alert['launch-image'] = self.launch_image
        return alert


@attr.s
class Notification:
    alert = attr.ib(validator=attr.validators.instance_of(Alert))
    badge = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(int))
    )
    sound = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    content_available = attr.ib(
        default=False,
        validator=attr.validators.instance_of(bool)
    )
    category = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    thread_id = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    extra = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(dict))
    )

    def get_dict(self) -> Dict[str, Any]:
        attr.validate(self)
        apns = {
            'alert': self.alert.get_dict()
        }
        if self.badge:
            apns['badge'] = self.badge
        if self.sound:
            apns['sound'] = self.sound
        if self.content_available:
            apns['content-available'] = 1
        if self.category:
            apns['category'] = self.category
        if self.thread_id:
            apns['thread-id'] = self.thread_id
        raw = {
            'aps': apns
        }
        if self.extra:
            raw.update(self.extra)
        return raw

    def encode(self) -> bytes:
        raw = self.get_dict()
        s = json.dumps(
            raw,
            ensure_ascii=False,
            separators=(',', ':'),
            sort_keys=True
        )
        return s.encode('utf-8')



from setuptools import setup, find_packages


setup(
    version='1.0.0.dev5',
    name='aapns',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    python_requires='>=3.6',
    install_requires=[
        'h2>3',
        'attrs',
        'structlog',
    ],
    extras_require={
        'cli': ['click']
    },
    entry_points={
        'console_scripts': [
            'aapns = aapns.cli:main'
        ]
    },
    license='APLv2',
    classifiers=[
        'Framework :: AsyncIO',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: Apache Software License'
    ]
)
